package ru.dbotthepony.mc.otm.container

import com.mojang.serialization.Codec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.core.HolderLookup
import net.minecraft.core.registries.BuiltInRegistries
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.NbtOps
import net.minecraft.nbt.Tag
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.neoforged.neoforge.common.util.INBTSerializable
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.core.isNotEmpty
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.data.codec.minRange
import java.util.function.Predicate

class SlottedContainer(
	slots: Collection<SlotProvider>,
	private val stillValid: Predicate<Player>,
	private val globalChangeListeners: Array<Runnable>
) : IAutomatedContainer, INBTSerializable<Tag> {
	private val slots: Array<ContainerSlot>

	init {
		val itr = slots.iterator()
		this.slots = Array(slots.size) { itr.next().create(this, it) }
	}

	override val hasFilterableSlots: Boolean = this.slots.any { it is IFilteredContainerSlot }
	private var suppressListeners = false

	override fun clearContent() {
		suppressListeners = true

		try {
			slots.forEach { it.remove() }
			notifyChanged()
		} finally {
			suppressListeners = false
		}
	}

	override fun containerSlot(slot: Int): IAutomatedContainerSlot {
		return slots[slot]
	}

	fun notifyChanged() {
		if (suppressListeners) return
		globalChangeListeners.forEach { it.run() }
	}

	// called by outside code (vanilla and other unaware mods)
	override fun setChanged() {
		suppressListeners = true
		var hasChanges = false

		try {
			slots.forEach { hasChanges = it.observeChanges() || hasChanges }
		} finally {
			suppressListeners = false

			if (hasChanges)
				notifyChanged()
		}
	}

	override fun getContainerSize(): Int {
		return slots.size
	}

	override fun stillValid(player: Player): Boolean {
		return stillValid.test(player)
	}

	private data class LegacySerializedItem(val item: ItemStack, val slot: Int) {
		companion object {
			val CODEC: Codec<LegacySerializedItem> = RecordCodecBuilder.create {
				it.group(
					ItemStack.OPTIONAL_CODEC.fieldOf("item").forGetter { it.item },
					Codec.INT.minRange(0).fieldOf("slot").forGetter { it.slot },
				).apply(it, ::LegacySerializedItem)
			}
		}
	}

	private data class LegacySerializedFilter(val item: Item, val slot: Int)  {
		companion object {
			val CODEC: Codec<LegacySerializedFilter> = RecordCodecBuilder.create {
				it.group(
					BuiltInRegistries.ITEM.byNameCodec().fieldOf("item").forGetter { it.item },
					Codec.INT.minRange(0).fieldOf("slot").forGetter { it.slot },
				).apply(it, ::LegacySerializedFilter)
			}
		}
	}

	private data class LegacySerializedState(
		val items: List<LegacySerializedItem>,
		val filters: List<LegacySerializedFilter>
	) {
		companion object {
			val CODEC: Codec<LegacySerializedState> = RecordCodecBuilder.create {
				it.group(
					Codec.list(LegacySerializedItem.CODEC).fieldOf("items").forGetter { it.items },
					Codec.list(LegacySerializedFilter.CODEC).fieldOf("filters").forGetter { it.filters },
				).apply(it, ::LegacySerializedState)
			}
		}
	}

	private val lostItems = ArrayList<CompoundTag>()
	private var provider: HolderLookup.Provider? = null

	fun takeLostItems(): List<ItemStack> {
		if (lostItems.isEmpty())
			return listOf()

		val provider = provider?.createSerializationContext(NbtOps.INSTANCE) ?: return listOf()
		val result = ArrayList<ItemStack>()

		for (entry in lostItems) {
			if ("item" in entry) {
				ItemStack.OPTIONAL_CODEC.decode(provider, entry["item"])
					.ifError { LOGGER.warn("Unable to deserialize 'lost' item: ${it.message()}") }
					.ifSuccess { if (it.first.isNotEmpty) result.add(it.first) }
			}
		}

		lostItems.clear()
		return result
	}

	fun takeLostItems(player: ServerPlayer) {
		for (stack in takeLostItems()) {
			player.inventory.add(stack)

			if (stack.isNotEmpty) {
				player.drop(stack, false, false)
			}
		}
	}

	override fun serializeNBT(provider: HolderLookup.Provider): ListTag {
		return ListTag().also {
			for (slot in slots) {
				it.add(slot.serializeNBT(provider))
			}

			it.addAll(lostItems)
		}
	}

	override fun deserializeNBT(provider: HolderLookup.Provider, nbt: Tag) {
		lostItems.clear()
		slots.forEach { it.clear() }

		if (nbt is CompoundTag) {
			// legacy container
			LegacySerializedState.CODEC.decode(provider.createSerializationContext(NbtOps.INSTANCE), nbt)
				.resultOrPartial { LOGGER.error("Error deserializing container: $it") }
				.ifPresent {
					val (items, filters) = it.first

					// excessive items will be lost
					for ((item, slot) in items) {
						if (slot in 0 until containerSize) {
							slots[slot].item = item
						} else if (item.isNotEmpty) {
							ItemStack.CODEC.encodeStart(provider.createSerializationContext(NbtOps.INSTANCE), item)
								.ifError { LOGGER.warn("Unable to serialize 'lost' item: ${it.message()}") }
								.ifSuccess { s ->
									this.provider = provider

									lostItems.add(CompoundTag().also {
										it["item"] = s
									})
								}
						}
					}

					for ((filter, slot) in filters) {
						if (slot in 0 until containerSize) {
							val getSlot = slots[slot]

							if (getSlot is IFilteredContainerSlot) {
								getSlot.filter = filter
							}
						}
					}
				}
		} else if (nbt is ListTag) {
			// normal container
			for ((i, element) in nbt.withIndex()) {
				if (element !is CompoundTag) {
					LOGGER.error("Deserializing mattery container: Expected compound tag at $i, got $element")
					continue
				}

				if (i in 0 until containerSize) {
					slots[i].deserializeNBT(provider, element)
				} else {
					lostItems.add(element)
					this.provider = provider
				}
			}
		} else {
			LOGGER.error("Unable to deserialize mattery container, expected CompoundTag or ListTag, got $nbt")
		}

		notifyChanged()
	}

	fun interface SlotProvider {
		fun create(container: SlottedContainer, index: Int): ContainerSlot
	}

	class Builder {
		private val slots = ArrayList<SlotProvider>()
		private var stillValid = Predicate<Player> { true }
		private val globalChangeListeners = ArrayList<Runnable>()

		fun add(slot: SlotProvider): Builder {
			slots.add(slot)
			return this
		}

		fun add(amount: Int, provider: SlotProvider): Builder {
			for (i in 0 until amount)
				slots.add(provider)

			return this
		}

		fun stillValid(predicate: Predicate<Player>): Builder {
			this.stillValid = predicate
			return this
		}

		fun onChanged(listener: Runnable): Builder {
			globalChangeListeners.add(listener)
			return this
		}

		fun copy(): Builder {
			val copy = Builder()
			copy.slots.addAll(slots)
			copy.globalChangeListeners.addAll(globalChangeListeners)
			copy.stillValid = stillValid
			return copy
		}

		fun build(): SlottedContainer {
			return SlottedContainer(slots, stillValid, globalChangeListeners.toTypedArray())
		}
	}


	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
